/*--------------------------------------------------------------------*\
  File:      AddOn.h
  Creator:   Matt Bogosian <mbogosian@usa.net>
  Copyright: (c)1998, Matt Bogosian. All rights reserved.
  Description: Header file describing the AddOn abstract base class.
  ID:        $Id: AddOn.h,v 1.3 1998/07/08 05:33:51 mattb Exp $
  Conventions:
      #defines - all uppercase letters with words separated by
          underscores.
          (E.G., #define MY_DEFINE 16).
      New data types (classes, structs, typedefs, etc.) - begin with
          an uppercase letter followed by lowercase words separated by
          uppercase letters.
          (E.G., typedef int MyTypedef;).
      Global constants (declared with const) - begin with "k_"
          followed by lowercase words separated by underscores.
          (E.G., const int k_my_constant(16);).
      Global variables - begin with "g_" followed by lowercase words
          separated by underscores.
          (E.G., int g_my_global;).
      Local variables - begin with a lowercase letter followed by
          lowercase words separated by underscores.
          (E.G., int my_local;).
      Argument variables - begin with "a_" followed by lowercase words
          separated by underscores.
          (E.G., ..., int a_my_argument, ...).
      Functions - begin with a lowercase letter followed by lowercase
          words separated by uppercase letters.
          (E.G., void myFunction(void);).
      Static member constants (declared with const) - begin with "mk_"
          followed by lowercase words separated by underscores.
          (E.G., static const int MyClass::mk_my_constant(16);).
      Static member variables - begin with "mg_" followed by lowercase
          words separated by underscores.
          (E.G., static int MyClass::mg_my_global(16);).
      Member variables - begin with "m_" followed by lowercase words
          separated by underscores.
          (E.G., int m_my_member;).
      Member Functions - begin with an uppercase letter followed by
          lowercase words separated by uppercase letters.
          (E.G., void MyClass::MyFunction(void);).
\*--------------------------------------------------------------------*/


/*--------------------------------------------------------------------*\
  =-=-=-=-=-=-=-=-=-=-=-=-=- Included Files -=-=-=-=-=-=-=-=-=-=-=-=-=
\*--------------------------------------------------------------------*/

#include "Arrayomatic.h"

#include <Messenger.h>
#include <OS.h>
#include <string.h>


#ifndef ADD_ON_H
#define ADD_ON_H


/*--------------------------------------------------------------------*\
  =-=-=-=-=-=-=-= Definitions, Enums, Typedefs, Consts =-=-=-=-=-=-=-=
\*--------------------------------------------------------------------*/

#ifndef DECL_SPEC

#if defined DLL_EXPORT_SYMBOLS
#define DECL_SPEC __declspec(dllexport)
#elif defined DLL_IMPORT_SYMBOLS
#define DECL_SPEC __declspec(dllimport)
#else
#define DECL_SPEC
#endif

#endif    // DECL_SPEC


/*--------------------------------------------------------------------*\
  =-=-=-=-=-=-=-=-=-=-=-=-= Structs, Classes =-=-=-=-=-=-=-=-=-=-=-=-=
\*--------------------------------------------------------------------*/

class Application;
class BLooper;
class KThread;

/*--------------------------------------------------------------------*\
  Struct name:      AddOnData
  Inherits from:    none
  New data members: public char m_name[B_FILE_NAME_LENGTH] - the name
                        of the add-on.
                    public Application *m_application - the
                        application with which the add-on should
                        interact.
  Description: Struct to contain the data necessary to instantiate an
      add-on.
\*--------------------------------------------------------------------*/

struct AddOnData
{
	char m_name[B_FILE_NAME_LENGTH];
	Application *m_application;
};

#ifdef DLL_EXPORT_SYMBOLS
#pragma export on
#endif

/*--------------------------------------------------------------------*\
  Class name:       AddOn
  Inherits from:    none
  New data members: private char m_name[B_FILE_NAME_LENGTH + 1] - the
                        name of the add-on.
                    private Application *m_application - the
                        application with which to interface.
                    private BMessenger *m_messanger - the messenger
                        for the handler to which to send the settings
                        messages.
                    private Arrayomatic<BLooper *> m_loopers - the
                       looper(s) which the add-on represents.
                    private Arrayomatic<KThread *> m_threads - the
                       thread(s) which the add-on represents.
                    private Arrayomatic<thread_id> m_thread_ids - the
                       thread ID(s) which the add-on represents.
  Description: Abstract base class to implement a generic independant
      thread-based add-on interface. This interface allows for
      notifications requiring the add-ons to quit smartly. It also
      allows for possible future interface expansion/enhancement.
      
      To use, simply derive a class from AddOn and define the
      appropriate hook functions:
      
        class MyAddOn : public AddOn
        {
          // Define the necessary hook functions here
        };
      
      Then, instantiate it and make sure it's valid:
      
        AddOn *add_on(new MyAddOn(...));
        
        if (add_on->Status() != B_NO_ERROR)
        {
          delete add_on;
        }
      
      Note: this class MUST be allocated dynamically (using the new
      operator). Also, the only time you should explicitly delete an
      add-on is if construction fails and you're still in the
      dvsrvrNewAddon() function. If dvsrvrNewAddon() returns a valid
      add-on, then the destructor is called for you you call
      RequestKill() (below) or after the program calls Quit() (below).
\*--------------------------------------------------------------------*/

class DECL_SPEC AddOn
{
	public:
	
		// Public member functions
		
/*--------------------------------------------------------------------*\
  Function name: AddOn
  Member of:     public AddOn
  Defined in:    AddOn.cpp
  Arguments:     const void * const a_data - the necessary basic data
                     for the add-on (this is given as an argument to
                     the dvsrvrNewAddOn() hook function. You must pass
                     that argument straight to the constructor for the
                     add-on. Do not pass go. Do not collect $200. (Bad
                     things may happen if the add-on is robbed of this
                     data.)
                 status_t * const a_err - a placeholder for the error
                     encountered during construction (see below).
                     Default: NULL.
  Throws:        none
  Description: Default class constructor. You must check the value of
      a_err or the return value of Status() after invoking the
      constructor to see if the object is usable. See the Status()
      function (below) for possible error values.
\*--------------------------------------------------------------------*/
		
		AddOn(const void * const a_data, status_t  * const a_err = NULL);
		
/*--------------------------------------------------------------------*\
  Function name: virtual ~AddOn
  Member of:     public AddOn
  Defined in:    AddOn.cpp
  Arguments:     none
  Throws:        none
  Description: Class destructor. The only time you should explicitly
      delete an add-on is if construction fails and you're still in
      the dvsrvrNewAddon() function. If dvsrvrNewAddon() returns a
      valid add-on, then the destructor is called for you you call
      RequestKill() (below) or after the program calls Quit() (below).
\*--------------------------------------------------------------------*/
		
		virtual ~AddOn(void);
		
/*--------------------------------------------------------------------*\
  Function name: inline AddLooper
  Member of:     public AddOn
  Arguments:     BLooper * const a_looper - the looper to add.
  Returns:       status_t - the error code associated with the
                     operation (see below).
  Throws:        none
  Description: Function to add a looper to the list of loopers that
      this add-on represents. Unless removed with the RemoveLooper()
      function (below), the given looper will automatically be Quit()
      when the add-on is deleted. Possible return values are as
      follows:

      B_NO_ERROR - success.
      B_NO_MEMORY - there was not enough memory to perform the
          operation.
\*--------------------------------------------------------------------*/
		
		inline status_t AddLooper(BLooper * const a_looper)
		{
			return m_loopers.PushItem(a_looper);
		}
		
/*--------------------------------------------------------------------*\
  Function name: inline AddThread
  Member of:     public AddOn
  Arguments:     KThread * const a_thread - the thread to add.
  Returns:       status_t - the error code associated with the
                     operation (see below).
  Throws:        none
  Description: Function to add a thread to the list of loopers that
      this add-on represents. Unless removed with the RemoveThread()
      function (below), the given thread will automatically be
      Kill()ed and deleted when the add-on is deleted. Possible return
      values are as follows:

      B_NO_ERROR - success.
      B_NO_MEMORY - there was not enough memory to perform the
          operation.
\*--------------------------------------------------------------------*/
		
		inline status_t AddThread(KThread * const a_thread)
		{
			return m_threads.PushItem(a_thread);
		}
		
/*--------------------------------------------------------------------*\
  Function name: inline AddThreadID
  Member of:     public AddOn
  Arguments:     const thread_id a_thread - the thread ID to add.
  Returns:       status_t - the error code associated with the
                     operation (see below).
  Throws:        none
  Description: Function to add a thread ID to the list of thread IDs
      that this add-on represents. Unless removed with the
      RemoveThreadID() function (below), the given thread ID will
      automatically be killed (via kill_thread()) when the add-on is
      deleted. Possible return values are as follows:

      B_NO_ERROR - success.
      B_NO_MEMORY - there was not enough memory to perform the
          operation.
\*--------------------------------------------------------------------*/
		
		inline status_t AddThreadID(const thread_id a_thread)
		{
			return m_thread_ids.PushItem(a_thread);
		}
		
/*--------------------------------------------------------------------*\
  Function name: inline Name
  Member of:     public AddOn
  Arguments:     char a_name[B_FILE_NAME_LENGTH] - a placeholder for
                     the name of the add-on.
  Returns:       none
  Throws:        none
  Description: Function to retrieve the name of the add-on.
\*--------------------------------------------------------------------*/
		
		inline void Name(char a_name[B_FILE_NAME_LENGTH])
		{
			strncpy(a_name, m_name, B_FILE_NAME_LENGTH);
		}
		
/*--------------------------------------------------------------------*\
  Function name: virtual Quit
  Member of:     public AddOn
  Defined in:    AddOn.cpp
  Arguments:     none
  Returns:       none
  Throws:        none
  Description: Hook function called to quit the add-on. When this
      function returns, the connection MUST be closed and all
      appropriate resources reclaimed: this add-on will be deleted
      after this function is called. If your add-on is still running
      after this function returns, you risk segmentation faults and
      other nastiness.
\*--------------------------------------------------------------------*/
		
		virtual void Quit(void);
		
/*--------------------------------------------------------------------*\
  Function name: inline RemoveLooper
  Member of:     public AddOn
  Arguments:     BLooper * const a_looper - the looper to remove.
  Returns:       bool - false if the item could not be removed (i.e.,
                     it could not be found), true otherwise.
  Throws:        none
  Description: Function to remove a looper from the list of loopers
      that this add-on represents. Once removed, the given looper will
      not automatically be Quit() when the add-on is deleted.
\*--------------------------------------------------------------------*/
		
		inline bool RemoveLooper(BLooper * const a_looper)
		{
			return m_loopers.Remove(a_looper);
		}
		
/*--------------------------------------------------------------------*\
  Function name: inline RemoveThread
  Member of:     public AddOn
  Arguments:     KThread * const a_thread - the thread to remove.
  Returns:       bool - false if the item could not be removed (i.e.,
                     it could not be found), true otherwise.
  Throws:        none
  Description: Function to remove a thread from the list of threads
      that this add-on represents. Once removed, the given thread will
      not automatically be Kill()ed and deleted when the add-on is
      deleted.
\*--------------------------------------------------------------------*/
		
		inline bool RemoveThread(KThread * const a_thread)
		{
			return m_threads.Remove(a_thread);
		}
		
/*--------------------------------------------------------------------*\
  Function name: inline RemoveThreadID
  Member of:     public AddOn
  Arguments:     const thread_id a_thread - the thread ID to remove.
  Returns:       bool - false if the item could not be removed (i.e.,
                     it could not be found), true otherwise.
  Throws:        none
  Description: Function to remove a thread ID from the list of thread
      IDs that this add-on represents. Once removed, the given thread
      ID will not automatically be killed (via kill_thread()) when the
      add-on is deleted.
\*--------------------------------------------------------------------*/
		
		inline bool RemoveThreadID(const thread_id a_thread)
		{
			return m_thread_ids.Remove(a_thread);
		}
		
/*--------------------------------------------------------------------*\
  Function name: RequestKill
  Member of:     public AddOn
  Defined in:    AddOn.cpp
  Arguments:     none
  Returns:       none
  Throws:        none
  Description: Function to request the clean up of the add-on after
      its execution is completed. This function will ask the
      dev_server to delete this add-on. This is safe to call from
      within the thread that the add-on represents.
\*--------------------------------------------------------------------*/
		
		void RequestKill(void);
		
/*--------------------------------------------------------------------*\
  Function name: SettingDelete
  Member of:     public AddOn
  Defined in:    AddOn.cpp
  Arguments:     const char * const a_name - the name of the setting
                     to delete.
  Returns:       status_t - the error code associated with the
                     operation.
  Throws:        none
  Description: Function to delete a setting with a given name.
\*--------------------------------------------------------------------*/
		
		status_t SettingDelete(const char * const a_name);
		
/*--------------------------------------------------------------------*\
  Function name: SettingRead
  Member of:     public AddOn
  Defined in:    AddOn.cpp
  Arguments:     const char * const a_name - the name of the setting
                     to read.
                 void * const a_data - a placeholder for the data
                     associated with the setting.
                 size_t * const a_data_len - the length of the buffer
                     in which to read the data and a placeholder for
                     the actual length.
                 type_code * const a_type - a placeholder for the type
                     of the data.
  Returns:       status_t - the error code associated with the
                     operation.
  Throws:        none
  Description: Function to read the data associated with the setting
      of a given name.
\*--------------------------------------------------------------------*/
		
		status_t SettingRead(const char * const a_name, void * const a_data, size_t * const a_data_len, type_code * const a_type);
		
/*--------------------------------------------------------------------*\
  Function name: SettingWrite
  Member of:     public AddOn
  Defined in:    AddOn.cpp
  Arguments:     const char * const a_name - the name of the setting
                     to write.
                 const void * const a_data - the data associated with
                     the setting.
                 const size_t a_data_len - the length of the data.
                 const type_code a_type - the type of the data.
  Returns:       status_t - the error code associated with the
                     operation.
  Throws:        none
  Description: Function to write the data associated with the setting
      of a given name.
\*--------------------------------------------------------------------*/
		
		status_t SettingWrite(const char * const a_name, const void * const a_data, const size_t a_data_len, const type_code a_type);
		
/*--------------------------------------------------------------------*\
  Function name: virtual Start
  Member of:     public AddOn
  Defined in:    AddOn.cpp
  Arguments:     const port_id a_send - the port on which messages
                     should be sent (the receiving port sent by the
                     DevCnnctn object in the driver).
                 const port_id a_recv - the port on which messages
                     should be received (the sending port sent by the
                     DevCnnctn object in the driver).
  Returns:       status_t - the error code associated with the
                     operation (see below).
  Throws:        none
  Description: Hook function defined by dev_server add-ons to complete
      a connection with the driver that initiated it. Connections
      between drivers and dev_server add-ons are made via a two-step
      process. The calling of this function is the second step. If the
      call to this function is successful (as indicated by a return
      code of B_NO_ERROR), then the connection between the driver and
      the dev_server add-on should be complete.
      
      This function should make note of the relevant given ports and
      start its own thread(s) in whatever form is appropriate (e.g.,
      BLooper, PortThread, etc.).
      
      The function should return B_NO_ERROR on success, or an
      appropriate error code on failure. This error code will be sent
      back to the driver. If the function returns anything other than
      B_NO_ERROR, then the dev_server will delete this add-on WITHOUT
      calling its Quit() member function. Make sure to allow for this
      and clean up all resources appropriately if an error occurs.
\*--------------------------------------------------------------------*/
		
		virtual status_t Start(const port_id a_send, const port_id a_recv) = NULL;
		
/*--------------------------------------------------------------------*\
  Function name: virtual Status const
  Member of:     public AddOn
  Defined in:    AddOn.cpp
  Arguments:     none
  Returns:       status_t - B_NO_ERROR if the object is valid, another
                     error code if not (see below).
  Throws:        none
  Description: Function to check if the object is valid. This function
      must be called after creating the object to insure that its
      instantiation was successful. This function should probably not
      be used once the thread is in motion. Classes who wish to
      override this function, should always call this function first
      to insure that the underlying structure is valid. Possible
      return values are as follows:
      
      B_NO_ERROR - the object was created successfully.
      B_NO_MEMORY - on construction, there was not enough memory
          available to create the object.
\*--------------------------------------------------------------------*/
		
		virtual status_t Status(void) const;
	
	private:
	
		// Private data members
		char m_name[B_FILE_NAME_LENGTH + 1];
		Application *m_application;
		BMessenger m_messenger;
		Arrayomatic<BLooper *> m_loopers;
		Arrayomatic<KThread *> m_threads;
		Arrayomatic<thread_id> m_thread_ids;
		
		// Private prohibitted member functions
		AddOn(const AddOn &a_obj);
		AddOn &operator=(const AddOn &a_obj);
};

#ifdef DLL_EXPORT_SYMBOLS
#pragma export off
#endif


#endif    // ADD_ON_H
