/*--------------------------------------------------------------------*\
  File:      PortThread.h
  Creator:   Matt Bogosian <mbogosian@usa.net>
  Copyright: (c)1998, Matt Bogosian. All rights reserved.
  Description: Header file describing the PortThread abstract class.
  ID:        $Id: PortThread.h,v 1.9 1998/07/01 05:16:16 mattb Exp $
  Conventions:
      #defines - all uppercase letters with words separated by
          underscores.
          (E.G., #define MY_DEFINE 16).
      New data types (classes, structs, typedefs, etc.) - begin with
          an uppercase letter followed by lowercase words separated by
          uppercase letters.
          (E.G., typedef int MyTypedef;).
      Global constants (declared with const) - begin with "k_"
          followed by lowercase words separated by underscores.
          (E.G., const int k_my_constant(16);).
      Global variables - begin with "g_" followed by lowercase words
          separated by underscores.
          (E.G., int g_my_global;).
      Local variables - begin with a lowercase letter followed by
          lowercase words separated by underscores.
          (E.G., int my_local;).
      Argument variables - begin with "a_" followed by lowercase words
          separated by underscores.
          (E.G., ..., int a_my_argument, ...).
      Functions - begin with a lowercase letter followed by lowercase
          words separated by uppercase letters.
          (E.G., void myFunction(void);).
      Static member constants (declared with const) - begin with "mk_"
          followed by lowercase words separated by underscores.
          (E.G., static const int MyClass::mk_my_constant(16);).
      Static member variables - begin with "mg_" followed by lowercase
          words separated by underscores.
          (E.G., static int MyClass::mg_my_global(16);).
      Member variables - begin with "m_" followed by lowercase words
          separated by underscores.
          (E.G., int m_my_member;).
      Member Functions - begin with an uppercase letter followed by
          lowercase words separated by uppercase letters.
          (E.G., void MyClass::MyFunction(void);).
\*--------------------------------------------------------------------*/


/*--------------------------------------------------------------------*\
  =-=-=-=-=-=-=-=-=-=-=-=-=- Included Files -=-=-=-=-=-=-=-=-=-=-=-=-=
\*--------------------------------------------------------------------*/

#include "KThread.h"


#ifndef PORT_THREAD_H
#define PORT_THREAD_H


/*--------------------------------------------------------------------*\
  =-=-=-=-=-=-=-= Definitions, Enums, Typedefs, Consts =-=-=-=-=-=-=-=
\*--------------------------------------------------------------------*/

#ifndef DECL_SPEC

#if defined DLL_EXPORT_SYMBOLS
#define DECL_SPEC __declspec(dllexport)
#elif defined DLL_IMPORT_SYMBOLS
#define DECL_SPEC __declspec(dllimport)
#else
#define DECL_SPEC
#endif

#endif    // DECL_SPEC


/*--------------------------------------------------------------------*\
  =-=-=-=-=-=-=-=-=-=-=-=-= Structs, Classes =-=-=-=-=-=-=-=-=-=-=-=-=
\*--------------------------------------------------------------------*/

#ifdef DLL_EXPORT_SYMBOLS
#pragma export on
#endif

/*--------------------------------------------------------------------*\
  Class name:       PortThread
  Inherits from:    public KThread
  Defined in:       PortThread.cpp
  New data members: private static const char
                        mk_default_name[B_OS_NAME_LENGTH] - the
                        default name of the thread.
                    protected size_t m_buf_size - the size (in bytes)
                        of the internal buffer.
                    protected uchar *m_buf - the internal buffer
                        provided as a matter of convenience for
                        classes derived from this one.
                    private port_id m_send - the port on which
                        messages will be sent.
                    private port_id m_recv - the port on which
                        messages will be received.
  Description: Abstract class to implement a connection thread for
      reading from and writing to its ports.
\*--------------------------------------------------------------------*/

#ifndef PortThread_SUPER
#define PortThread_SUPER KThread
#endif

class DECL_SPEC PortThread : public PortThread_SUPER
{
	private:
	
		// Private static const data members
		static const char mk_default_name[B_OS_NAME_LENGTH];
	
	public:
	
		// Public member functions
		
/*--------------------------------------------------------------------*\
  Function name: PortThread
  Member of:     public PortThread
  Defined in:    PortThread.cpp
  Arguments:     const port_id a_send - the port on which messages
                     will be sent. Default: B_BAD_PORT_ID.
                 const port_id a_recv - the port on which messages
                     will be received (this could be the same as the
                     sending port in certain cases). Default:
                     B_BAD_PORT_ID.
                 const size_t a_buf_size - the size of the internal
                     buffer. Default: 0.
                 const char a_name[B_OS_NAME_LENGTH] - the name of the
                     thread. Default: mk_default_name.
                 const int32 a_priority - the priority of the thread.
                     Default: B_NORMAL_PRIORITY.
                 status_t * const a_err - a placeholder for the error
                     encountered during construction (see below).
                     Default: NULL.
  Throws:        none
  Description: Default class constructor. You must check the value of
      a_err or the return value of Status() after invoking the
      constructor to see if the object is usable. See the Status()
      function (below) for possible error values.
\*--------------------------------------------------------------------*/
		
		PortThread(const port_id a_send = B_BAD_PORT_ID, const port_id a_recv = B_BAD_PORT_ID, const size_t a_buf_size = 0, const char a_name[B_OS_NAME_LENGTH] = mk_default_name, const int32 a_priority = B_NORMAL_PRIORITY, status_t * const a_err = NULL);
		
/*--------------------------------------------------------------------*\
  Function name: virtual ~PortThread
  Member of:     public PortThread
  Defined in:    PortThread.cpp
  Arguments:     none
  Throws:        none
  Description: Class destructor.
\*--------------------------------------------------------------------*/
		
		virtual ~PortThread(void);
		
/*--------------------------------------------------------------------*\
  Function name: inline ReadSend
  Member of:     public PortThread
  Arguments:     int32 &a_msg_code - a placeholder for the message
                     code.
                 void * const a_buf - a placeholder for the message.
                     Default: NULL.
                 const size_t a_buf_size - the size (in bytes) of the
                     message buffer. Default: 0.
                 const uint32 a_flags - the flags. Default: 0.
                 const bigtime_t a_timeout - the timeout
                     (in microseconds). Default: B_INFINITE_TIMEOUT.
  Returns:       ssize_t - the number of bytes read on success, an
                     error code on failure (see below).
  Throws:        none
  Description: Function to read a message from the port on which
      messages are sent. Note: this call is NOT guarded by a
      semaphore. To insure exclusive access, Lock() the thread first.
      Possible return values are as follows:
      
      B_BAD_PORT_ID - the port is no longer valid.
      B_WOULD_BLOCK - a timeout of 0 was specified and the port was
          not available.
      B_TIMED_OUT - the timeout was reached and a message could not be
          be read.
\*--------------------------------------------------------------------*/
		
		inline ssize_t ReadSend(int32 &a_msg_code, void * const a_buf = NULL, const size_t a_buf_size = 0, const uint32 a_flags = 0, const bigtime_t a_timeout = B_INFINITE_TIMEOUT)
		{
			return Read(m_send, a_msg_code, a_buf, a_buf_size, a_flags, a_timeout);
		}
		
/*--------------------------------------------------------------------*\
  Function name: inline ReceiveBufferSize
  Member of:     public PortThread
  Arguments:     const uint32 a_flags - the flags. Default: 0.
                 const bigtime_t a_timeout - the timeout
                     (in microseconds). Default: B_INFINITE_TIMEOUT.
  Returns:       ssize_t - the buffer size on success, an error code
                     on failure (see below).
  Throws:        none
  Description: Function to report the receiving port's buffer size.
      Note: this call is NOT guarded by a semaphore. To insure
      exclusive access, Lock() the thread first. Possible return
      values are as follows:
      
      B_BAD_PORT_ID - the port is no longer valid.
      B_WOULD_BLOCK - a timeout of 0 was specified and the port was
          not available.
      B_TIMED_OUT - the timeout was reached and the buffer size could
          not be read.
\*--------------------------------------------------------------------*/
		
		inline ssize_t ReceiveBufferSize(const uint32 a_flags = 0, const bigtime_t a_timeout = B_INFINITE_TIMEOUT)
		{
			return BufferSize(m_recv, a_flags, a_timeout);
		}
		
/*--------------------------------------------------------------------*\
  Function name: inline SendBufferSize
  Member of:     public PortThread
  Arguments:     const uint32 a_flags - the flags. Default: 0.
                 const bigtime_t a_timeout - the timeout
                     (in microseconds). Default: B_INFINITE_TIMEOUT.
  Returns:       ssize_t - the buffer size on success, an error code
                     on failure (see below).
  Throws:        none
  Description: Function to report the sending port's buffer size.
      Note: this call is NOT guarded by a semaphore. To insure
      exclusive access, Lock() the thread first. Possible return
      values are as follows:
      
      B_BAD_PORT_ID - the port is no longer valid.
      B_WOULD_BLOCK - a timeout of 0 was specified and the port was
          not available.
      B_TIMED_OUT - the timeout was reached and the buffer size could
          not be read.
\*--------------------------------------------------------------------*/
		
		inline ssize_t SendBufferSize(const uint32 a_flags = 0, const bigtime_t a_timeout = B_INFINITE_TIMEOUT)
		{
			return BufferSize(m_send, a_flags, a_timeout);
		}
		
/*--------------------------------------------------------------------*\
  Function name: virtual Status const
  Member of:     public PortThread
  Defined in:    PortThread.cpp
  Arguments:     none
  Returns:       status_t - B_NO_ERROR if the object is valid, another
                     error code if not (see below).
  Throws:        none
  Description: Derived hook function to check if the object is valid.
      Possible return values (in addition to those from the function
      from which this is derived) are as follows:
      
      B_BAD_PORT_ID - on construction, neither of the object's ports
           were found to be valid.
\*--------------------------------------------------------------------*/
		
		virtual status_t Status(void) const;
		
/*--------------------------------------------------------------------*\
  Function name: inline WriteReceive
  Member of:     public PortThread
  Arguments:     const int32 a_msg_code - the message code.
                 const void * const a_msg - the message. Default:
                     NULL.
                 const size_t a_msg_len - the length (in bytes) of the
                     message. Default: 0.
                 const uint32 a_flags - the flags. Default: 0.
                 const bigtime_t a_timeout - the timeout
                     (in microseconds). Default: B_INFINITE_TIMEOUT.
  Returns:       status_t - the error code associated with the
                     operation (see below).
  Throws:        none
  Description: Function to write a message to the port on which
      messages are received. Note: this call is NOT guarded by a
      semaphore. To insure exclusive access, Lock() the thread first.
      Possible return values are as follows:
      
      B_NO_ERROR - success.
      B_BAD_PORT_ID - the port is no longer valid.
      B_WOULD_BLOCK - a timeout of 0 was specified and the port was
          not available.
      B_TIMED_OUT - the timeout was reached and the message could not
          be sent.
\*--------------------------------------------------------------------*/
		
		inline status_t WriteReceive(const int32 a_msg_code, const void * const a_msg = NULL, const size_t a_msg_len = 0, const uint32 a_flags = 0, const bigtime_t a_timeout = B_INFINITE_TIMEOUT)
		{
			return Write(m_recv, a_msg_code, a_msg, a_msg_len, a_flags, a_timeout);
		}
	
	protected:
	
		// Protected data members
		size_t m_buf_size;
		uchar *m_buf;
		
		// Protected member functions
		
/*--------------------------------------------------------------------*\
  Function name: inline ReadReceive
  Member of:     protected PortThread
  Arguments:     int32 &a_msg_code - a placeholder for the message
                     code.
                 void * const a_buf - a placeholder for the message.
                     Default: NULL.
                 const size_t a_buf_size - the size (in bytes) of the
                     message buffer. Default: 0.
                 const uint32 a_flags - the flags. Default: 0.
                 const bigtime_t a_timeout - the timeout
                     (in microseconds). Default: B_INFINITE_TIMEOUT.
  Returns:       ssize_t - the number of bytes read on success, an
                     error code on failure (see below).
  Throws:        none
  Description: Function to read a message from the port on which
      messages are received. Note: this call is NOT guarded by a
      semaphore. To insure exclusive access, Lock() the thread first.
      Possible return values are as follows:
      
      B_BAD_PORT_ID - the port is no longer valid.
      B_WOULD_BLOCK - a timeout of 0 was specified and the port was
          not available.
      B_TIMED_OUT - the timeout was reached and a message could not be
          be read.
\*--------------------------------------------------------------------*/
		
		inline ssize_t ReadReceive(int32 &a_msg_code, void * const a_buf = NULL, const size_t a_buf_size = 0, const uint32 a_flags = 0, const bigtime_t a_timeout = B_INFINITE_TIMEOUT)
		{
			return Read(m_recv, a_msg_code, a_buf, a_buf_size, a_flags, a_timeout);
		}
		
/*--------------------------------------------------------------------*\
  Function name: inline WriteSend
  Member of:     protected PortThread
  Arguments:     const int32 a_msg_code - the message code.
                 const void * const a_msg - the message. Default:
                     NULL.
                 const size_t a_msg_len - the length (in bytes) of the
                     message. Default: 0.
                 const uint32 a_flags - the flags. Default: 0.
                 const bigtime_t a_timeout - the timeout
                     (in microseconds). Default: B_INFINITE_TIMEOUT.
  Returns:       status_t - the error code associated with the
                     operation (see below).
  Throws:        none
  Description: Function to write a message to the port on which
      messages are sent. Note: this call is NOT guarded by a
      semaphore. To insure exclusive access, Lock() the thread first.
      Possible return values are as follows:
      
      B_NO_ERROR - success.
      B_BAD_PORT_ID - the port is no longer valid.
      B_WOULD_BLOCK - a timeout of 0 was specified and the port was
          not available.
      B_TIMED_OUT - the timeout was reached and the message could not
          be sent.
\*--------------------------------------------------------------------*/
		
		inline status_t WriteSend(const int32 a_msg_code, const void * const a_msg = NULL, const size_t a_msg_len = 0, const uint32 a_flags = 0, const bigtime_t a_timeout = B_INFINITE_TIMEOUT)
		{
			return Write(m_send, a_msg_code, a_msg, a_msg_len, a_flags, a_timeout);
		}
	
	private:
	
		// Private data members
		port_id m_send;
		port_id m_recv;
		
		// Private static member functions
		
/*--------------------------------------------------------------------*\
  Function name: static BufferSize
  Member of:     private PortThread
  Arguments:     port_id &a_port - the port of which to get the buffer
                     size.
                 const uint32 a_flags - the flags. Default: 0.
                 const bigtime_t a_timeout - the timeout
                     (in microseconds). Default: B_INFINITE_TIMEOUT.
  Returns:       ssize_t - the buffer size on success, an error code
                     on failure (see below).
  Throws:        none
  Description: Function to report the a given port's buffer size.
      Possible return values are as follows:
      
      B_BAD_PORT_ID - the port is no longer valid.
      B_WOULD_BLOCK - a timeout of 0 was specified and the port was
          not available.
      B_TIMED_OUT - the timeout was reached and the buffer size could
          not be read.
\*--------------------------------------------------------------------*/
		
		static ssize_t BufferSize(port_id &a_port, const uint32 a_flags = 0, const bigtime_t a_timeout = B_INFINITE_TIMEOUT);
		
/*--------------------------------------------------------------------*\
  Function name: static Read
  Member of:     private PortThread
  Defined in:    PortThread.cpp
  Arguments:     port_id &a_port - the port from which to read the
                     message.
                 int32 &a_msg_code - a placeholder for the message
                     code.
                 void * const a_buf - a placeholder for the message.
                     Default: NULL.
                 const size_t a_buf_size - the size (in bytes) of the
                     message buffer. Default: 0.
                 const uint32 a_flags - the flags. Default: 0.
                 const bigtime_t a_timeout - the timeout
                     (in microseconds). Default: B_INFINITE_TIMEOUT.
  Returns:       ssize_t - the number of bytes read on success, an
                     error code on failure (see below).
  Throws:        none
  Description: Function to read a message from a given port. Possible
      return values are as follows:
      
      B_BAD_PORT_ID - the port is no longer valid.
      B_WOULD_BLOCK - a timeout of 0 was specified and the port was
          not available.
      B_TIMED_OUT - the timeout was reached and a message could not be
          be read.
\*--------------------------------------------------------------------*/
		
		static ssize_t Read(port_id &a_port, int32 &a_msg_code, void * const a_buf = NULL, const size_t a_buf_size = 0, const uint32 a_flags = 0, const bigtime_t a_timeout = B_INFINITE_TIMEOUT);
		
/*--------------------------------------------------------------------*\
  Function name: static Write
  Member of:     private PortThread
  Defined in:    PortThread.cpp
  Arguments:     port_id &a_port - the port to which to write the
                     message.
                 const int32 a_msg_code - the message code.
                 const void * const a_msg - the message. Default:
                     NULL.
                 const size_t a_msg_len - the length (in bytes) of the
                     message. Default: 0.
                 const uint32 a_flags - the flags. Default: 0.
                 const bigtime_t a_timeout - the timeout
                     (in microseconds). Default: B_INFINITE_TIMEOUT.
  Returns:       status_t - the error code associated with the
                     operation (see below).
  Throws:        none
  Description: Function to write a message to a given port. Possible
      return values are as follows:
      
      B_NO_ERROR - success.
      B_BAD_PORT_ID - the port is no longer valid.
      B_WOULD_BLOCK - a timeout of 0 was specified and the port was
          not available.
      B_TIMED_OUT - the timeout was reached and the message could not
          be sent.
\*--------------------------------------------------------------------*/
		
		static status_t Write(port_id &a_port, const int32 a_msg_code, const void * const a_msg = NULL, const size_t a_msg_len = 0, const uint32 a_flags = 0, const bigtime_t a_timeout = B_INFINITE_TIMEOUT);
};

#ifdef DLL_EXPORT_SYMBOLS
#pragma export off
#endif


#endif    // PORT_THREAD_H
