/*--------------------------------------------------------------------*\
  File:      thruput.cpp
  Creator:   Matt Bogosian <mbogosian@usa.net>
  Copyright: (c)1998, Matt Bogosian. All rights reserved.
  Description: Source file containing the hook functions for the
      kernel-loadable device driver as defined in Drivers.h. This may
      look a little like C++, but don't use any classes; the new and
      delete operators are not to be found here.
  ID:        $Id: thruput.cpp,v 1.5 1998/07/03 12:51:26 mattb Exp $
  Conventions:
      #defines - all uppercase letters with words separated by
          underscores.
          (E.G., #define MY_DEFINE 16).
      New data types (classes, structs, typedefs, etc.) - begin with
          an uppercase letter followed by lowercase words separated by
          uppercase letters.
          (E.G., typedef int MyTypedef;).
      Global constants (declared with const) - begin with "k_"
          followed by lowercase words separated by underscores.
          (E.G., const int k_my_constant(16);).
      Global variables - begin with "g_" followed by lowercase words
          separated by underscores.
          (E.G., int g_my_global;).
      Local variables - begin with a lowercase letter followed by
          lowercase words separated by underscores.
          (E.G., int my_local;).
      Argument variables - begin with "a_" followed by lowercase words
          separated by underscores.
          (E.G., ..., int a_my_argument, ...).
      Functions - begin with a lowercase letter followed by lowercase
          words separated by uppercase letters.
          (E.G., void myFunction(void);).
      Static member constants (declared with const) - begin with "mk_"
          followed by lowercase words separated by underscores.
          (E.G., static const int MyClass::mk_my_constant(16);).
      Static member variables - begin with "mg_" followed by lowercase
          words separated by underscores.
          (E.G., static int MyClass::mg_my_global(16);).
      Member variables - begin with "m_" followed by lowercase words
          separated by underscores.
          (E.G., int m_my_member;).
      Member Functions - begin with an uppercase letter followed by
          lowercase words separated by uppercase letters.
          (E.G., void MyClass::MyFunction(void);).
\*--------------------------------------------------------------------*/


/*--------------------------------------------------------------------*\
  =-=-=-=-=-=-=-=-=-=-=-=-=- Included Files -=-=-=-=-=-=-=-=-=-=-=-=-=
\*--------------------------------------------------------------------*/

#include "DevCnnctn.h"
#include <AppDefs.h>
#include <KernelExport.h>
#include <Drivers.h>
#include <errno.h>
#include <fcntl.h>
#include <malloc.h>
#include <string.h>
#include <unistd.h>


/*--------------------------------------------------------------------*\
  =-=-=-=-=-=-=-= Definitions, Enums, Typedefs, Consts =-=-=-=-=-=-=-=
\*--------------------------------------------------------------------*/

// This is returned by publish_devices() to indicate to the kernel
// which devices this driver defines
static const char * const k_dev_names[] =
{
	"ports/thruput",
	NULL
};


/*--------------------------------------------------------------------*\
  =-=-=-=-=-=-=-=-=-=-=-=-= Structs, Classes =-=-=-=-=-=-=-=-=-=-=-=-=
\*--------------------------------------------------------------------*/

struct ThruputCookie
{
	DevCnnctn m_connection;
	int m_fd;
};


/*--------------------------------------------------------------------*\
  =-=-=-=-=-=-=-=-=-=-=-=- Function Prototypes =-=-=-=-=-=-=-=-=-=-=-=
\*--------------------------------------------------------------------*/

/*--------------------------------------------------------------------*\
  Function name: static thruputClose
  Arguments:     void * const a_cookie - the cookie associated with
                     this call.
  Returns:       status_t - B_NO_ERROR on success, an error code on
                     failure.
  Throws:        none
  Description: Hook function to handle the closing of the device.
\*--------------------------------------------------------------------*/

static status_t thruputClose(void * const a_cookie);

/*--------------------------------------------------------------------*\
  Function name: static thruputCtrl
  Arguments:     void * const a_cookie - the cookie associated with
                     this call.
                 const uint32 a_op - the control command.
                 const void * const a_data - the data associated with
                     the command.
                 const size_t a_data_len - the length (in bytes) of
                     a_data.
  Returns:       status_t - B_NO_ERROR on success, an error code on
                     failure.
  Throws:        none
  Description: Hook function to handle calls the controlling (e.g.,
      via ioctl()) of the device.
\*--------------------------------------------------------------------*/

static status_t thruputCtrl(void * const a_cookie, const uint32 a_op, const void * const a_data, const size_t a_data_len);

/*--------------------------------------------------------------------*\
  Function name: static thruputFree
  Arguments:     void * const a_cookie - the cookie associated with
                     this call.
  Returns:       status_t - B_NO_ERROR on success, an error code on
                     failure.
  Throws:        none
  Description: Hook function to handle the freeing of the device.
\*--------------------------------------------------------------------*/

static status_t thruputFree(void * const a_cookie);

/*--------------------------------------------------------------------*\
  Function name: static thruputOpen
  Arguments:     const char * const a_name - the name of the device to
                     open. Some device drivers can define several
                     devices; this let's you know which one is being
                     opened.
                 const uint32 a_flags - the flags (e.g., O_RDONLY)
                     sent by the call to open().
                 const void ** const a_cookie - a placeholder for the
                     a cookie to be associated with further calls to
                     the device. This is passed back to the device as
                     the first argument of all other device hook
                     functions.
  Returns:       status_t - B_NO_ERROR on success, an error code on
                     failure.
  Throws:        none
  Description: Hook function to handle the opening of the device.
\*--------------------------------------------------------------------*/

static status_t thruputOpen(const char * const a_name, const uint32 a_flags, const void ** const a_cookie);

/*--------------------------------------------------------------------*\
  Function name: static thruputRead
  Arguments:     void * const a_cookie - the cookie associated with
                     this call.
                 const off_t a_pos - the position from which to read.
                 void * const a_data - a placeholder for the data to
                     be read.
                 size_t * const a_data_len - the length (in bytes) of
                     a_data, or the maximum amount of data that should
                     be read from the device. The actual amount read
                     should be placed in this variable before this
                     function returns.
  Returns:       status_t - B_NO_ERROR on success, an error code on
                     failure.
  Throws:        none
  Description: Hook function to handle reading from the device.
\*--------------------------------------------------------------------*/

static status_t thruputRead(void * const a_cookie, const off_t a_pos, void * const a_data, size_t * const a_data_len);

/*--------------------------------------------------------------------*\
  Function name: static thruputWrite
  Arguments:     void * const a_cookie - the cookie associated with
                     this call.
                 const off_t a_pos - the position to which to write.
                 void * const a_data - the data to be written.
                 size_t * const a_data_len - the length (in bytes) of
                     a_data, or the maximum amount of data that should
                     be written to the device. The actual amount
                     written should be placed in this variable before
                     this function returns.
  Returns:       status_t - B_NO_ERROR on success, an error code on
                     failure.
  Throws:        none
  Description: Hook function to handle writing to the device.
\*--------------------------------------------------------------------*/

static status_t thruputWrite(void * const a_cookie, const off_t a_pos, const void * const a_data, size_t * const a_data_len);


/*--------------------------------------------------------------------*\
  =-=-=-=-=-=-=-=-=-=-=-=-= Global Variables =-=-=-=-=-=-=-=-=-=-=-=-=
\*--------------------------------------------------------------------*/

// This struct is returned by find_device() when called with the
// appropriate device name
static device_hooks g_dev_hooks =
{
	reinterpret_cast<device_open_hook>(&thruputOpen),
	reinterpret_cast<device_close_hook>(&thruputClose),
	reinterpret_cast<device_free_hook>(&thruputFree),
	reinterpret_cast<device_control_hook>(&thruputCtrl),
	reinterpret_cast<device_read_hook>(&thruputRead),
	reinterpret_cast<device_write_hook>(&thruputWrite)
};

/*--------------------------------------------------------------------*\
  =-=-=-=-=-=-=-=-=-=-=-= Function Definitions =-=-=-=-=-=-=-=-=-=-=-=
\*--------------------------------------------------------------------*/

//====================================================================
static status_t thruputClose(void * const a_cookie)
//====================================================================
{
	ThruputCookie *thruput(static_cast<ThruputCookie *>(a_cookie));
	
	// Get rid of the file descriptor
	int fd(thruput->m_fd);
	thruput->m_fd = -1;
	
	// Try to close the real file
	return (close(fd) < 1) ? errno : B_NO_ERROR;
}

//====================================================================
static status_t thruputCtrl(void * const a_cookie, const uint32 a_op, const void * const a_data, const size_t a_data_len)
//====================================================================
{
	ThruputCookie *thruput(static_cast<ThruputCookie *>(a_cookie));
	
	// Try to send the ioctl to the real file
	int ret_val(ioctl(thruput->m_fd, a_op, a_data, a_data_len));
	
	return (ret_val < 0) ? errno : ret_val;
}

//====================================================================
static status_t thruputFree(void * const a_cookie)
//====================================================================
{
	ThruputCookie *thruput(static_cast<ThruputCookie *>(a_cookie));
	
	// Break the connection with the dev_server and free our
	// memory like a good boy
	DevCnnctnDestroy(&thruput->m_connection);
	free(thruput);
	
	return B_NO_ERROR;
}

//====================================================================
static status_t thruputOpen(const char * const a_name, const uint32 a_flags, const void ** const a_cookie)
//====================================================================
{
	ThruputCookie *thruput;
	
	// Check for an allocation error
	if ((thruput = static_cast<ThruputCookie *>(malloc(sizeof (ThruputCookie)))) == NULL)
	{
		return B_NO_MEMORY;
	}
	
	// Try to open the connection
	status_t err;
	port_id port(create_port(1, ""));
	const char *name(strrchr(a_name, '/'));
	name = (name == NULL) ? a_name : &name[1];
	
	if ((err = DevCnnctnCreate(&thruput->m_connection, name, port, port)) != B_NO_ERROR)
	{
		delete_port(port);
		free(thruput);
		return err;
	}
	
	// This shouldn't loop more than once, but just in case...try
	// to get the real file we're supposed to open
	ssize_t msg_len;
	int32 msg_code;
	uchar buf[B_PATH_NAME_LENGTH];
	
	do
	{
		msg_len = DevCnnctnRead(thruput->m_connection, &msg_code, buf, B_PATH_NAME_LENGTH);
	}
	while (msg_len >= B_NO_ERROR
		&& msg_code != B_REFS_RECEIVED);
	
	// Check for an error
	if (msg_len < B_NO_ERROR)
	{
		DevCnnctnDestroy(&thruput->m_connection);
		free(thruput);
		return msg_len;
	}
	
	// Try to open the real file
	if ((thruput->m_fd = open(reinterpret_cast<char *>(buf), a_flags)) < 0)
	{
		DevCnnctnDestroy(&thruput->m_connection);
		free(thruput);
		return errno;
	}
	
	// Write the first thruput message to the dev_server
	*a_cookie = thruput;
	uint64 bps_read(0), bps_write(0);
	bigtime_t time(0);
	memcpy(&buf[0], &bps_read, sizeof (bps_read));
	memcpy(&buf[sizeof (bps_read)], &bps_write, sizeof (bps_write));
	memcpy(&buf[sizeof (bps_read) + sizeof (bps_write)], &time, sizeof (time));
	DevCnnctnWrite(thruput->m_connection, B_VALUE_CHANGED, buf, sizeof (bps_read) + sizeof (bps_write) + sizeof (time));
	
	return B_NO_ERROR;
}

//====================================================================
static status_t thruputRead(void * const a_cookie, const off_t a_pos, void * const a_data, size_t * const a_data_len)
//====================================================================
{
	ThruputCookie *thruput(static_cast<ThruputCookie *>(a_cookie));
	
	// Try to read from the real file
	ssize_t bytes_read(read_pos(thruput->m_fd, a_pos, a_data, *a_data_len));
	
	// Check for an error
	if (bytes_read >= B_NO_ERROR
		&& thruput->m_connection != NULL)
	{
		// Try to read the thruput message from the dev_server
		int32 msg_code;
		uint64 bps_read, bps_write;
		uchar buf[sizeof (bps_read) + sizeof (bps_write) + sizeof (bigtime_t)];
		
		if (DevCnnctnRead(thruput->m_connection, &msg_code, buf, sizeof (bps_read) + sizeof (bps_write) + sizeof (bigtime_t)) < B_NO_ERROR)
		{
			DevCnnctnDestroy(&thruput->m_connection);
		}
		else
		{
			// Modify it and write it back
			memcpy(&bps_read, &buf[0], sizeof (bps_read));
			bps_read += bytes_read * CHAR_BIT;
			memcpy(&buf[0], &bps_read, sizeof (bps_read));
			DevCnnctnWrite(thruput->m_connection, B_VALUE_CHANGED, buf, sizeof (bps_read) + sizeof (bps_write) + sizeof (bigtime_t));
		}
		
		*a_data_len = bytes_read;
		bytes_read = B_NO_ERROR;
	}
	
	return bytes_read;
}

//====================================================================
static status_t thruputWrite(void * const a_cookie, const off_t a_pos, const void * const a_data, size_t * const a_data_len)
//====================================================================
{
	ThruputCookie *thruput(static_cast<ThruputCookie *>(a_cookie));
	
	// Try to write from the real file
	ssize_t bytes_written(write_pos(thruput->m_fd, a_pos, a_data, *a_data_len));
	
	// Check for an error
	if (bytes_written >= B_NO_ERROR
		&& thruput->m_connection != NULL)
	{
		// Try to read the thruput message from the dev_server
		int32 msg_code;
		uint64 bps_read, bps_write;
		uchar buf[sizeof (bps_read) + sizeof (bps_write) + sizeof (bigtime_t)];
		
		if (DevCnnctnRead(thruput->m_connection, &msg_code, buf, sizeof (bps_read) + sizeof (bps_write) + sizeof (bigtime_t)) < B_NO_ERROR)
		{
			DevCnnctnDestroy(&thruput->m_connection);
		}
		else
		{
			// Modify it and write it back
			memcpy(&bps_write, &buf[sizeof (bps_read)], sizeof (bps_write));
			bps_write += bytes_written * CHAR_BIT;
			memcpy(&buf[sizeof (bps_read)], &bps_write, sizeof (bps_write));
			DevCnnctnWrite(thruput->m_connection, B_VALUE_CHANGED, buf, sizeof (bps_read) + sizeof (bps_write) + sizeof (bigtime_t));
		}
		
		*a_data_len = bytes_written;
		bytes_written = B_NO_ERROR;
	}
	
	return bytes_written;
}

#pragma export on

//====================================================================
extern "C" device_hooks *find_device(const char * const a_name)
//====================================================================
{
	if (!strcmp (a_name, k_dev_names[0]))
	{
		return &g_dev_hooks;
	}
	
	return NULL;
}

//====================================================================
extern "C" status_t init_driver(void)
//====================================================================
{
	return B_NO_ERROR;
}

//====================================================================
extern "C" const char **publish_devices(void)
//====================================================================
{
	return k_dev_names;
}

//====================================================================
extern "C" void uninit_driver(void)
//====================================================================
{
	;
}

#pragma export off
